@@ -64,7 +64,6 @@ static	void	hfree	(struct Hist *);
 
 /* #define DEBUG_HIST 1 */
 
-static const int fastMergeErase = 1;
 static unsigned histCount = 0;		/* number elements on history list */
 static int histlen = 0;
 static struct Hist *histTail = NULL;     /* last element on history list */
@@ -205,44 +204,67 @@ addWordToHash(struct hashValue *h, const Char *word)
 {
     uint32_t a = h->a, b = h->b, c = h->c;
 #ifdef SHORT_STRINGS
+#define GETK    if ((k = (uChar)*word++) == 0) break
 #ifdef WIDE_STRINGS
     assert(sizeof(Char) >= 4);
     while (1) {
 	unsigned k;
-	if ((k = (uChar)*word++) == 0) break; a += k;
-	if ((k = (uChar)*word++) == 0) break; b += k;
-	if ((k = (uChar)*word++) == 0) break; c += k;
+	GETK;
+	a += k;
+	GETK;
+	b += k;
+	GETK;
+	c += k;
 	mix(a, b, c);
     }
 #else
     assert(sizeof(Char) == 2);
     while (1) {
 	unsigned k;
-	if ((k = (uChar)*word++) == 0) break; a += k;
-	if ((k = (uChar)*word++) == 0) break; a += k << 16;
-	if ((k = (uChar)*word++) == 0) break; b += k;
-	if ((k = (uChar)*word++) == 0) break; b += k << 16;
-	if ((k = (uChar)*word++) == 0) break; c += k;
-	if ((k = (uChar)*word++) == 0) break; c += k << 16;
+	GETK;
+	a += k;
+	GETK;
+	a += k << 16;
+	GETK;
+	b += k;
+	GETK;
+	b += k << 16;
+	GETK;
+	c += k;
+	GETK;
+	c += k << 16;
 	mix(a, b, c);
     }
 #endif
 #else
     assert(sizeof(Char) == 1);
+#define GETK    if ((k = *word++) == 0) break
     while (1) {
 	unsigned k;
-	if ((k = *word++) == 0) break; a += k;
-	if ((k = *word++) == 0) break; a += k << 8;
-	if ((k = *word++) == 0) break; a += k << 16;
-	if ((k = *word++) == 0) break; a += k << 24;
-	if ((k = *word++) == 0) break; b += k;
-	if ((k = *word++) == 0) break; b += k << 8;
-	if ((k = *word++) == 0) break; b += k << 16;
-	if ((k = *word++) == 0) break; b += k << 24;
-	if ((k = *word++) == 0) break; c += k;
-	if ((k = *word++) == 0) break; c += k << 8;
-	if ((k = *word++) == 0) break; c += k << 16;
-	if ((k = *word++) == 0) break; c += k << 24;
+	GETK;
+	a += k;
+	GETK;
+	a += k << 8;
+	GETK;
+	a += k << 16;
+	GETK;
+	a += k << 24;
+	GETK;
+	b += k;
+	GETK;
+	b += k << 8;
+	GETK;
+	b += k << 16;
+	GETK;
+	b += k << 24;
+	GETK;
+	c += k;
+	GETK;
+	c += k << 8;
+	GETK;
+	c += k << 16;
+	GETK;
+	c += k << 24;
 	mix(a, b, c);
     }
 #endif
@@ -603,15 +625,6 @@ heq(const struct wordent *a0, const struct wordent *b0)
     }
 }
 
-/* Renumber entries following p, which we will be deleting. */
-PG_STATIC void
-renumberHist(struct Hist *p)
-{
-    int n = p->Href;
-    while ((p = p->Hnext))
-        p->Href = n--;
-}
-
 /* The hash table is implemented as an array of pointers to Hist entries.  Each
  * entry is located in the table using hash2tableIndex() and checking the
  * following entries in case of a collision (linear rehash).  Free entries in
@@ -941,8 +954,6 @@ enthist(
                  * to insert the new entry, then remember the place. */
                 if (mflg && Htime != 0 && p->Hprev->Htime >= Htime)
                     pTime = p->Hprev;
-		if (!fastMergeErase)
-		    renumberHist(p);	/* Reset Href of subsequent entries */
                 hremove(p);
 		hfree(p);
                 p = NULL;               /* so new entry is allocated below */
@@ -1008,11 +1019,10 @@ enthist(
               }
           }
         /* pp is now the last entry with time >= to np. */
-	if (!fastMergeErase) {		/* renumber at end of loadhist */
-	    /* Before inserting np after pp, bubble its Hnum & Href values down
-	     * through the earlier part of list. */
-	    bubbleHnumHrefDown(np, pp);
-	}
+	/* renumber at end of loadhist */
+	/* Before inserting np after pp, bubble its Hnum & Href values down
+	 * through the earlier part of list. */
+	bubbleHnumHrefDown(np, pp);
     }
     else
         pp = &Histlist;                 /* insert at beginning of history */
@@ -1054,15 +1064,15 @@ phist(struct Hist *hp, int hflg)
         output_raw = 1;
 	cleanup_push(&old_output_raw, output_raw_restore);
 	if (hflg & HIST_TIME)
-	    /* 
+	    /*
 	     * Make file entry with history time in format:
-	     * "+NNNNNNNNNN" (10 digits, left padded with ascii '0') 
+	     * "+NNNNNNNNNN" (10 digits, left padded with ascii '0')
 	     */
 
 	    xprintf("#+%010lu\n", (unsigned long)hp->Htime);
 
 	if (HistLit && hp->histline)
-	    xprintf("%S\n", hp->histline);
+	    xprintf("%" TCSH_S "\n", hp->histline);
 	else
 	    prlex(&hp->Hlex);
         cleanup_until(&old_output_raw);
@@ -1174,6 +1184,17 @@ dohist(Char **vp, struct command *c)
     }
 }
 
+void
+cleanhist(void)
+{
+    struct Hist *hp, *np;
+
+    for (hp = &Histlist; (np = hp->Hnext) != NULL;) {
+	if (np->Hnum != HIST_PURGE)
+	    return;
+	hremove(np), hfree(np);
+    }
+}
 
 char *
 fmthist(int fmt, ptr_t ptr)
@@ -1186,7 +1207,7 @@ fmthist(int fmt, ptr_t ptr)
 	return xasprintf("%6d", hp->Hnum);
     case 'R':
 	if (HistLit && hp->histline)
-	    return xasprintf("%S", hp->histline);
+	    return xasprintf("%" TCSH_S, hp->histline);
 	else {
 	    Char *istr, *ip;
 	    char *p;
@@ -1216,17 +1237,23 @@ dotlock_cleanup(void* lockpath)
 
 /* Save history before exiting the shell. */
 void
-rechist(Char *fname, int ref)
+rechist(Char *xfname, int ref)
 {
     Char    *snum, *rs;
-    int     fp, ftmp, oldidfds;
+    int     fp, ftmp, oldidfds, ophup_disabled;
     struct varent *shist;
     char path[MAXPATHLEN];
     struct stat st;
+    static Char *fname;
     static Char   *dumphist[] = {STRhistory, STRmhT, 0, 0};
 
-    if (fname == NULL && !ref) 
+    if (xfname == NULL && !ref)
 	return;
+
+    fname = xfname;
+    ophup_disabled = phup_disabled;
+    phup_disabled = 1;
+
     /*
      * If $savehist is just set, we use the value of $history
      * else we use the value in $savehist
@@ -1249,8 +1276,8 @@ rechist(Char *fname, int ref)
     /*
      * The 'savehist merge' feature is intended for an environment
      * with numerous shells being in simultaneous use. Imagine
-     * any kind of window system. All these shells 'share' the same 
-     * ~/.history file for recording their command line history. 
+     * any kind of window system. All these shells 'share' the same
+     * ~/.history file for recording their command line history.
      * We try to handle the case of multiple shells trying to merge
      * histories at the same time, by creating semi-unique filenames
      * and saving the history there first and then trying to rename
@@ -1261,7 +1288,7 @@ rechist(Char *fname, int ref)
 		 * by optional lock parameter to savehist.
      *
      * jw.
-     */ 
+     */
     /*
      * We need the didfds stuff before loadhist otherwise
      * exec in a script will fail to print if merge is set.
@@ -1281,6 +1308,7 @@ rechist(Char *fname, int ref)
 	}
 
 	if (merge) {
+	    jmp_buf_t osetexit;
 	    if (lock) {
 #ifndef WINNT_NATIVE
 		char *lockpath = strsave(short2str(fname));
@@ -1290,17 +1318,21 @@ rechist(Char *fname, int ref)
 		    cleanup_push(lockpath, dotlock_cleanup);
 #endif
 	    }
-	    loadhist(fname, 1);
+	    getexit(osetexit);
+	    if (setexit() == 0)
+		loadhist(fname, 1);
+	    resexit(osetexit);
 	}
     }
     rs = randsuf();
-    xsnprintf(path, sizeof(path), "%S.%S", fname, rs);
+    xsnprintf(path, sizeof(path), "%" TCSH_S ".%" TCSH_S, fname, rs);
     xfree(rs);
 
     fp = xcreat(path, 0600);
     if (fp == -1) {
 	didfds = oldidfds;
 	cleanup_until(fname);
+	phup_disabled = ophup_disabled;
 	return;
     }
     /* Try to preserve ownership and permissions of the original history file */
@@ -1322,9 +1354,10 @@ rechist(Char *fname, int ref)
 #ifndef WINNT_NATIVE
     (void)rename(path, short2str(fname));
 #else
-    (void)ReplaceFile( short2str(fname),path,NULL,0,NULL,NULL);
+    (void)ReplaceFile(short2str(fname), path, NULL, 0, NULL, NULL);
 #endif
     cleanup_until(fname);
+    phup_disabled = ophup_disabled;
 }
 
 
@@ -1343,16 +1376,6 @@ loadhist(Char *fname, int mflg)
 	loadhist_cmd[2] = STRtildothist;
 
     dosource(loadhist_cmd, NULL);
-
-    /* During history merging (enthist sees mflg set), we disable management of
-     * Hnum and Href (because fastMergeErase is true).  So now reset all the
-     * values based on the final ordering of the history list. */
-    if (mflg) {
-	int n = eventno;
-        struct Hist *hp = &Histlist;
-        while ((hp = hp->Hnext))
-	    hp->Hnum = hp->Href = n--;
-    }
 }
 
 void
